==> pygwalker-0.4.5a2.tar.gz: pygwalker-0.4.5a2/app/src/utils/communication.tsx <==
import { v4 as uuidv4 } from 'uuid';
import commonStore from '../store/common';

interface IResponse {
    data?: any;
    message?: string;
    code: number;
}

interface ICommunication {
    sendMsg: (action: string, data: any, timeout?: number) => Promise<IResponse>;
    registerEndpoint: (action: string, callback: (data: any) => any) => void;
    sendMsgAsync: (action: string, data: any, rid: string | null) => void;
}

const getSignalName = (rid: string) => {
    return `hacker-comm-pyg-done-signal-${rid}`;
}

const getCurrentJupyterEnv = () => {
    const host = window.parent.location.host;
    if (host === 'datalore.jetbrains.com') {
        return 'datalore';
    }
    return "jupyter";
}

const raiseRequestError = (message: string, code: number) => {
    let showMsg = (
        <>{message}</>
    )
    switch (code) {
        case 20001:
            showMsg = (
                <>
                <p className="py-1">kanaries_api_key is not valid.</p>
                <p className="py-1">Please set kanaries_api_key first.</p>
                <p className="py-1">If you are not kanaries user, please register it from
                    <a className="font-semibold px-1" href="https://kanaries.net/home/access" target='_blank'>
                        https://kanaries.net/home/access
                    </a>
                </p>
                <p className="py-1">Then refer
                    <a className="font-semibold px-1" href="https://github.com/Kanaries/pygwalker/wiki/How-to-get-api-key-of-kanaries%3F" target='_blank'>
                        document
                    </a>
                    to set kanaries_api_key.</p>
                </>
            );
            break;
        case 20002:
            showMsg = (<>
                <p className="py-1">
                    The usage of cloud config has reached the upper limit.
                    If you want to use more cloud config files, please subscribe to different cloud packages according to your usage:
                    <a className="font-semibold px-1" href="https://kanaries.net/home/pygwalker" target='_blank'>
                        https://kanaries.net/home/pygwalker
                    </a>
                </p>
            </>)
            break;
    }
    commonStore.setNotification({
        type: "error",
        title: "Error",
        message: showMsg || "Unknown error",
    }, 20_000);
}

const initJupyterCommunication = (gid: string) => {
    const kernelTextCount = 5;
    let curKernelTextIndex = 0;
    const jupyterEnv = getCurrentJupyterEnv();
    const document = window.parent.document;
    const htmlText = document.getElementsByClassName(`hacker-comm-pyg-html-store-${gid}`)[0].childNodes[1] as HTMLInputElement;
    const kernelTextList = Array.from({ length: kernelTextCount }, (_, index) => index).map(index => {
        return document.getElementsByClassName(`hacker-comm-pyg-kernel-store-${gid}-${index}`)[0].childNodes[1] as HTMLInputElement;
    })

    const requestTask = [] as any[];
    const endpoints = new Map<string, (data: any) => any>();
    const bufferMap = new Map<string, any>();

    const onMessage = (msg: string) => {
        const data = JSON.parse(msg);
        const action = data.action;
        if (action === "finish_request") {
            bufferMap.set(data.rid, data.data);
            document.dispatchEvent(new CustomEvent(getSignalName(data.rid)));
            return
        }
        if (action === "finish_batch_request") {
            data.data.forEach((resp: any) => {
                bufferMap.set(resp.rid, resp.data);
                document.dispatchEvent(new CustomEvent(getSignalName(resp.rid)));
            })
            return
        }
        const callback = endpoints.get(action);
        if (callback) {
            const resp = callback(data.data) ?? {};
            sendMsgAsync("finish_request", resp, data.rid);
        }
    }

    const sendMsg = async(action: string, data: any, timeout: number = 30_000) => {
        const rid = uuidv4();
        const promise = new Promise<any>((resolve, reject) => {
            setTimeout(() => {
                sendMsgAsync(action, data, rid);
            }, 0);
            const timer = setTimeout(() => {
                raiseRequestError("communication timeout", 0);
                reject(new Error("get result timeout"));
            }, timeout);
            document.addEventListener(getSignalName(rid), (_) => {
                clearTimeout(timer);
                const resp = bufferMap.get(rid);
                if (resp.code !== 0) {
                    raiseRequestError(resp.message, resp.code);
                    reject(new Error(resp.message));
                }
                resolve(resp);
            });
        });

        return promise;
    }

    const sendMsgAsync = (action: string, data: any, rid: string | null) => {
        rid = rid ?? uuidv4();
        requestTask.push({ action, data, rid });
        if (requestTask.length === 1) {
            setTimeout(() => {
                batchSendMsgAsync(requestTask.splice(0, requestTask.length));
            }, 300);
        }
    }

    const batchSendMsgAsync = (data: any) => {
        const rid = uuidv4();
        const event = new Event("input", { bubbles: true })
        const kernelText = kernelTextList[curKernelTextIndex];
        kernelText.value = JSON.stringify({ gid: gid, rid: rid, action: "batch_request", data });
        kernelText.dispatchEvent(event);
        curKernelTextIndex = (curKernelTextIndex + 1) % kernelTextCount;
    }

    const registerEndpoint = (action: string, callback: (data: any) => any) => {
        endpoints.set(action, callback);
    }

    if (jupyterEnv === "datalore") {
        const kernel = (window.parent as any).Jupyter.notebook.kernel;
        if (kernel.__pre_can_handle_message === undefined) {
            kernel.__pre_can_handle_message = kernel._can_handle_message;
        }
        kernel._can_handle_message = (msg: any) => {
            if (msg.msg_type === "comm_msg" && msg.content.data.method === "update") {
                const pygMsgStr = msg.content.data.state.value;
                try {
                    if (JSON.parse(pygMsgStr).gid === gid) {
                        onMessage(pygMsgStr);
                    }
                } catch (_) {}
            }
            return kernel.__pre_can_handle_message(msg);
        }
    } else {
        const observer = new MutationObserver((mutations) => {
            mutations.forEach((mutation) => {
                if (mutation.type === "attributes") {
                    onMessage(htmlText.value)
                }
            })
        })
        observer.observe(htmlText, {
            attributes: true,
            attributeFilter: ["placeholder"],
        })
    }

    return {
        sendMsg,
        registerEndpoint,
        sendMsgAsync,
    }
}

const initHttpCommunication = (gid: string, baseUrl: string) => {
    // temporary solution in streamlit could
    const domain = window.parent.document.location.host.split(".").slice(-2).join('.');
    let url = "";
    if (domain === "streamlit.app") {
        url = `/~/+/_stcore/_pygwalker/comm/${gid}`
    } else {
        url = `/${baseUrl}/${gid}`
    }

    const sendMsg = async(action: string, data: any, timeout: number = 30_000) => {
        const timer = setTimeout(() => {
            raiseRequestError("communication timeout", 0);
            throw(new Error("get result timeout"));
        }, timeout);
        try {
            const resp = await (await sendMsgAsync(action, data)).json();
            if (resp.code !== 0) {
                raiseRequestError(resp.message, resp.code);
                throw new Error(resp.message);
            }
            return resp;
        } finally {
            clearTimeout(timer);
        }
    }

    const sendMsgAsync = (action: string, data: any) => {
        const rid = uuidv4();
        return fetch(
            url,
            {
                method: "POST",
                headers: { "Content-Type": "application/json" },
                body: JSON.stringify({ gid, rid, action, data }),
            }
        )
    }

    const registerEndpoint = (_: string, __: (data: any) => any) => {}

    return {
        sendMsg,
        registerEndpoint,
        sendMsgAsync,
    }
}

export type { ICommunication };
export { initJupyterCommunication, initHttpCommunication };

==> pygwalker-0.4.7a1.tar.gz: pygwalker-0.4.7a1/app/src/utils/communication.tsx <==
import { v4 as uuidv4 } from 'uuid';
import commonStore from '../store/common';

interface IResponse {
    data?: any;
    message?: string;
    code: number;
}

interface ICommunication {
    sendMsg: (action: string, data: any, timeout?: number) => Promise<IResponse>;
    registerEndpoint: (action: string, callback: (data: any) => any) => void;
    sendMsgAsync: (action: string, data: any, rid: string | null) => void;
}

const getSignalName = (rid: string) => {
    return `hacker-comm-pyg-done-signal-${rid}`;
}

const getCurrentJupyterEnv = () => {
    const host = window.parent.location.host;
    if (host === 'datalore.jetbrains.com') {
        return 'datalore';
    }
    return "jupyter";
}

const raiseRequestError = (message: string, code: number) => {
    let showMsg = (
        <>{message}</>
    )
    switch (code) {
        case 20001:
            showMsg = (
                <>
                <p className="py-1">kanaries_api_key is not valid.</p>
                <p className="py-1">Please set kanaries_api_key first.</p>
                <p className="py-1">If you are not kanaries user, please register it from
                    <a className="font-semibold px-1" href="https://kanaries.net/home/access" target='_blank'>
                        https://kanaries.net/home/access
                    </a>
                </p>
                <p className="py-1">Then refer
                    <a className="font-semibold px-1" href="https://github.com/Kanaries/pygwalker/wiki/How-to-get-api-key-of-kanaries%3F" target='_blank'>
                        document
                    </a>
                    to set kanaries_api_key.</p>
                </>
            );
            break;
        case 20002:
            showMsg = (<>
                <p className="py-1">
                    The usage of cloud config has reached the upper limit.
                    If you want to use more cloud config files, please subscribe to different cloud packages according to your usage:
                    <a className="font-semibold px-1" href="https://kanaries.net/home/pygwalker" target='_blank'>
                        https://kanaries.net/home/pygwalker
                    </a>
                </p>
            </>)
            break;
    }
    commonStore.setNotification({
        type: "error",
        title: "Error",
        message: showMsg || "Unknown error",
    }, 20_000);
}

const initJupyterCommunication = (gid: string) => {
    const kernelTextCount = 5;
    let curKernelTextIndex = 0;
    const jupyterEnv = getCurrentJupyterEnv();
    const document = window.parent.document;
    const htmlText = document.getElementsByClassName(`hacker-comm-pyg-html-store-${gid}`)[0].childNodes[1] as HTMLInputElement;
    const kernelTextList = Array.from({ length: kernelTextCount }, (_, index) => index).map(index => {
        return document.getElementsByClassName(`hacker-comm-pyg-kernel-store-${gid}-${index}`)[0].childNodes[1] as HTMLInputElement;
    })

    const requestTask = [] as any[];
    const endpoints = new Map<string, (data: any) => any>();
    const bufferMap = new Map<string, any>();

    const onMessage = (msg: string) => {
        const data = JSON.parse(msg);
        const action = data.action;
        if (action === "finish_request") {
            bufferMap.set(data.rid, data.data);
            document.dispatchEvent(new CustomEvent(getSignalName(data.rid)));
            return
        }
        if (action === "finish_batch_request") {
            data.data.forEach((resp: any) => {
                bufferMap.set(resp.rid, resp.data);
                document.dispatchEvent(new CustomEvent(getSignalName(resp.rid)));
            })
            return
        }
        const callback = endpoints.get(action);
        if (callback) {
            const resp = callback(data.data) ?? {};
            sendMsgAsync("finish_request", resp, data.rid);
        }
    }

    const sendMsg = async(action: string, data: any, timeout: number = 30_000) => {
        const rid = uuidv4();
        const promise = new Promise<any>((resolve, reject) => {
            setTimeout(() => {
                sendMsgAsync(action, data, rid);
            }, 0);
            const timer = setTimeout(() => {
                raiseRequestError("communication timeout", 0);
                reject(new Error("get result timeout"));
            }, timeout);
            document.addEventListener(getSignalName(rid), (_) => {
                clearTimeout(timer);
                const resp = bufferMap.get(rid);
                if (resp.code !== 0) {
                    raiseRequestError(resp.message, resp.code);
                    reject(new Error(resp.message));
                }
                resolve(resp);
            });
        });

        return promise;
    }

    const sendMsgAsync = (action: string, data: any, rid: string | null) => {
        rid = rid ?? uuidv4();
        requestTask.push({ action, data, rid });
        if (requestTask.length === 1) {
            setTimeout(() => {
                batchSendMsgAsync(requestTask.splice(0, requestTask.length));
            }, 100);
        }
    }

    const batchSendMsgAsync = (data: any) => {
        const rid = uuidv4();
        const event = new Event("input", { bubbles: true })
        const kernelText = kernelTextList[curKernelTextIndex];
        kernelText.value = JSON.stringify({ gid: gid, rid: rid, action: "batch_request", data });
        kernelText.dispatchEvent(event);
        curKernelTextIndex = (curKernelTextIndex + 1) % kernelTextCount;
    }

    const registerEndpoint = (action: string, callback: (data: any) => any) => {
        endpoints.set(action, callback);
    }

    if (jupyterEnv === "datalore") {
        const kernel = (window.parent as any).Jupyter.notebook.kernel;
        if (kernel.__pre_can_handle_message === undefined) {
            kernel.__pre_can_handle_message = kernel._can_handle_message;
        }
        kernel._can_handle_message = (msg: any) => {
            if (msg.msg_type === "comm_msg" && msg.content.data.method === "update") {
                const pygMsgStr = msg.content.data.state.value;
                try {
                    if (JSON.parse(pygMsgStr).gid === gid) {
                        onMessage(pygMsgStr);
                    }
                } catch (_) {}
            }
            return kernel.__pre_can_handle_message(msg);
        }
    } else {
        const observer = new MutationObserver((mutations) => {
            mutations.forEach((mutation) => {
                if (mutation.type === "attributes") {
                    onMessage(htmlText.value)
                }
            })
        })
        observer.observe(htmlText, {
            attributes: true,
            attributeFilter: ["placeholder"],
        })
    }

    return {
        sendMsg,
        registerEndpoint,
        sendMsgAsync,
    }
}

interface IHttpRequestTask {
    data: {
        action: string;
        data: any;
        rid: string;
        gid: string;
    };
    resolve: (resp: any) => void;
    reject: (err: any) => void;
}

const initHttpCommunication = (gid: string, baseUrl: string) => {
    // temporary solution in streamlit could
    const domain = window.parent.document.location.host.split(".").slice(-2).join('.');
    let url = "";
    if (domain === "streamlit.app") {
        url = `/~/+/_stcore/_pygwalker/comm/${gid}`
    } else {
        url = `/${baseUrl}/${gid}`
    }

    const requestTask = [] as IHttpRequestTask[];

    const sendMsg = async(action: string, data: any, timeout: number = 30_000) => {
        const timer = setTimeout(() => {
            raiseRequestError("communication timeout", 0);
            throw(new Error("get result timeout"));
        }, timeout);
        try {
            const resp = await sendMsgAsync(action, data);
            if (resp.code !== 0) {
                raiseRequestError(resp.message, resp.code);
                throw new Error(resp.message);
            }
            return resp;
        } finally {
            clearTimeout(timer);
        }
    }

    const sendMsgAsync = (action: string, data: any) => {
        const rid = uuidv4();
        const promise = new Promise<any>((resolve, reject) => {
            requestTask.push({
                data: { action, data, rid, gid },
                resolve,
                reject,
            });
            if (requestTask.length === 1) {
                setTimeout(() => {
                    batchSendMsgAsync(requestTask.splice(0, requestTask.length));
                }, 100);
            }
        });
        return promise;
    }

    const batchSendMsgAsync = async(taskList: IHttpRequestTask[]) => {
        try {
            const resp = await fetch(
                url,
                {
                    method: "POST",
                    headers: { "Content-Type": "application/json" },
                    body: JSON.stringify({
                        gid: gid,
                        rid: uuidv4(),
                        action: "batch_request",
                        data: taskList.map(task => task.data)
                    }),
                }
            )
            const respJson = await resp.json();
            taskList.forEach((task, index) => {
                const taskResp = respJson[index];
                task.resolve(taskResp);
            })
        } catch (err) {
            taskList.forEach(task => {
                task.reject(err);
            })
        }
    }

    const registerEndpoint = (_: string, __: (data: any) => any) => {}

    return {
        sendMsg,
        registerEndpoint,
        sendMsgAsync,
    }
}

export type { ICommunication };
export { initJupyterCommunication, initHttpCommunication };
